//
//  STelSock	(SSL Telnet Socket)
//
//	Copyright (C) 1996-1999 NEC Corporation
//	All rights reserved.
//  Authos(s): Mizukoshi, Yasuhiro <mizukosi@ccs.mt.nec.co.jp>
//

static char CopyRightID[] = "Copyright(C) 1996-1999 NEC Corporation. All rights reserved.";

#ifdef _SSLEAY06_
static char ObjectID[] = "STelSock for SSLeay Ver.0.6.x";
#else
static char ObjectID[] = "STelSock for SSLeay Ver.0.8.x";
#endif

//	    $Archive: /sotools/stelsock/stelsock.c $
//			$Revision: 23 $
//			$Date: 99/03/01 5:17p $
static char VssID[] = "$Header: /sotools/stelsock/stelsock.c 23    99/03/01 5:17p Mizukosi $";

//
//	Comment:
//		This DLL makes it possible to communicate network applications
//		with SSL servers.
//

//
//	Functions:
//		SSLtelnet_connect
//		SSLtelnet_send
//		SSLftp_accept
//		SSLftp_send
//		accept
//		closesocket
//		connect
//		recv
//		send
//		socket
//		WSAAsyncSelect
//		WSAStartup
//		Set_SSLCTX
//		getkeycb
//		PassphraseDlgProc
//		Verify_CB
//		ErrorMsgBox
//		ErrmsgDlgProc
//		GetOptions
//		DllMain
//		LoadWinsock
//		FreeWinsock
//

#include	<windows.h>
#include	<winsock.h>
#include	<stdio.h>

#include	"ssl.h"
#include	"stelsock.h"
#include	"auth.h"
#include	"resource.h"

#define		GLOBALS
#include	"globals.h"


///////////////////////////////////////////////////////////////////
//
//	SSLtelnet_connect
//		This connect() is implemented for the TELNET applications.
//
int PASCAL FAR EXPORT
SSLtelnet_connect (SOCKET s, const struct sockaddr FAR *name, int namelen)
{
	int		ret;
	struct SocketList	*psl;
	unsigned long nbmode;

	psl = SocketListGetPos(s);

x_OutputDebugString("connect ");

	// Cancel async message temporally
	if (psl->hAsyncWnd){
		gft.WSAAsyncSelect(s, psl->hAsyncWnd, 0, 0);
	}

	// change to blocking-mode
	// In blocking-mode, it is easy to detect connect() error
	// and easy to write SSLtelnet negotiation code.
	if (psl->bNonBlocking){
		nbmode = FALSE;
		gft.ioctlsocket(s, FIONBIO, &nbmode);
	}

	// Connect to telnet server
	ret = gft.connect(s, name, namelen);

	// SSL telnet negotiation
	if (!ret && psl &&
		(giNegotiation == SSLONLY || DoSSLTelnet(s)) ){

#ifdef _SSLEAY06_
		psl->STelChk.ctx = SSL_CTX_new();
#else
		psl->STelChk.meth = SSLv23_method();
		SSLeay_add_ssl_algorithms();
		psl->STelChk.ctx = SSL_CTX_new(psl->STelChk.meth);
#endif /* _SSLEAY06_ */

		Set_SSLCTX(psl->STelChk.ctx);
		psl->STelChk.con = (SSL *)SSL_new(psl->STelChk.ctx);

		SSL_set_fd(psl->STelChk.con, s);

		if (SSL_connect(psl->STelChk.con) <= 0){
x_OutputDebugString("\n# SSL Not Active. #####\n");
			ErrorMsgBox("Error: SSL not active");

		} else {
			BOOL	bOpt;

x_OutputDebugString("\n# SSL Active. #####\n");

			psl->STelChk.bSSLActive = TRUE;

			bOpt = TRUE;
			if (gft.setsockopt(s, SOL_SOCKET, SO_OOBINLINE, (char *)&bOpt, sizeof(bOpt))
								== SOCKET_ERROR){
x_OutputDebugString("setsockopt error! ");
			}

		}

	}

	// If FD_CONNECT event is set, FD_CONNECT is posted here.
	if (psl && psl->hAsyncWnd && (psl->lAsyncEvent & FD_CONNECT) == FD_CONNECT){
		if (ret != SOCKET_ERROR){
			PostMessage(psl->hAsyncWnd, psl->uAsyncMsg, s, FD_CONNECT);
		} else {
			PostMessage(psl->hAsyncWnd, psl->uAsyncMsg, s, MAKELPARAM(FD_CONNECT, WSAGetLastError()));
		}
	}

	if (psl->bNonBlocking){
		nbmode = TRUE;
		gft.ioctlsocket(s, FIONBIO, &nbmode);
	}

	return ret;
}


/////////////////////////////////////////////////////////////////
//
//	SSLtelnet_send
//		This send() is implemented for the TELNET applications
//
int PASCAL FAR EXPORT
SSLtelnet_send (SOCKET s, const char FAR * buf, int len,
    int flags)
{
	int		ret;
	struct SocketList	*psl;

	psl = SocketListGetPos(s);

x_OutputDebugString("SSLtelnet_send ");

	if (psl && psl->STelChk.bSSLActive && !psl->STelChk.bInSSLCall){
x_OutputDebugString("S ");

		psl->STelChk.bInSSLCall = TRUE;
		ret = SSL_write(psl->STelChk.con, (char *)buf, len);
		psl->STelChk.bInSSLCall = FALSE;

	} else {
x_OutputDebugString("O ");

		ret = gft.send(s, buf, len, flags);

	}	

	return ret;
}


//														SSLtelnet
///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
//	SSLftp


///////////////////////////////////////////////////////////////////
//
//	SSLftp_accept
//		This accept() is implemented for the data connection by
//		the PORT command of FTP applications.
//
SOCKET PASCAL FAR EXPORT
SSLftp_accept (SOCKET s, struct sockaddr FAR *addr, int FAR *addrlen)
{
	SOCKET	ret;
	struct SocketList *pslCtrl, *pslData;

	pslCtrl = SocketListGetPos(gsFtpCtrl);

x_OutputDebugString("SSLftp_accept ");

	// Free Prev Data Connection
	pslData = SocketListGetPos(gsFtpData);
	if (pslData){
		if (pslData->STelChk.con){
			pslData->STelChk.bInSSLCall = TRUE;
			pslData->STelChk.bSSLActive = FALSE;
			SSL_free(pslData->STelChk.con);
			pslData->STelChk.con = NULL;
			pslData->STelChk.bInSSLCall = FALSE;
		}
		SocketListDelete(pslData->s);
	}

	// At first, accept.
	ret = gft.accept(s, addr, addrlen);

	if (ret != SOCKET_ERROR){

		SocketListAddTail(ret);
		gsFtpData = ret;

		if (pslCtrl && pslCtrl->STelChk.bSSLActive){

			pslData = SocketListGetPos(ret);
			pslData->STelChk.con = (SSL *)SSL_new(pslCtrl->STelChk.ctx);

			SSL_set_fd(pslData->STelChk.con, ret);

            SSL_set_verify(pslData->STelChk.con,
						   gVerifyDepth ? SSL_VERIFY_PEER : SSL_VERIFY_NONE, Verify_CB);
            SSL_copy_session_id(pslData->STelChk.con, pslCtrl->STelChk.con);

			if (SSL_connect(pslData->STelChk.con) <= 0){
x_OutputDebugString("\n# Data SSL Not Active. #####\n");
				ErrorMsgBox("Error: SSL not active");

			} else {
x_OutputDebugString("\n# Data SSL Active. #####\n");
				pslData->STelChk.bSSLActive = TRUE;
			}
		}
	}

	return ret;
}


/////////////////////////////////////////////////////////////////
//
//	SSLftp_send
//		This send() is implemented for the FTP applications.
//		This function negoticates with SSL ftpd.
//
int PASCAL FAR EXPORT
SSLftp_send (SOCKET s, const char FAR * buf, int len,
    int flags)
{
	int		ret;
	struct SocketList	*psl;
	unsigned long nbmode;

	psl = SocketListGetPos(s);

x_OutputDebugString("SSLftp_send ");

	// At first, STelSock negotiates with FTP server.
	// In other words, STelSock sends "SSL AUTH" to FTP server,
	// before FTP application sends "USER" command.
	if (gsFtpCtrl == s && gbFtpChecked == FALSE){
		gbFtpChecked = TRUE;

		// Cancel async message temporally
		if (psl->hAsyncWnd)
			gft.WSAAsyncSelect(s, psl->hAsyncWnd, 0, 0);

		// change to blocking-mode
		// In blocking-mode, it is easy to write SSLftp negotiation code.
		if (psl->bNonBlocking){
			nbmode = FALSE;
			gft.ioctlsocket(s, FIONBIO, &nbmode);
		}

		// Negotiate with FTP server.
		if (DoSSLFtp(s)){
#ifdef _SSLEAY06_
			psl->STelChk.ctx = SSL_CTX_new();
#else
			psl->STelChk.meth = SSLv23_method();
			SSLeay_add_ssl_algorithms();
			psl->STelChk.ctx = SSL_CTX_new(psl->STelChk.meth);
#endif /* _SSLEAY06_ */

			Set_SSLCTX(psl->STelChk.ctx);
			psl->STelChk.con = (SSL *)SSL_new(psl->STelChk.ctx);

			SSL_set_fd(psl->STelChk.con, s);

			if (SSL_connect(psl->STelChk.con) <= 0){
x_OutputDebugString("\n# SSL Not Active. #####\n");
				ErrorMsgBox("Error: SSL not active");

			} else {
x_OutputDebugString("\n# SSL Active. #####\n");
				// bSSLActive flag is turned on.
				psl->STelChk.bSSLActive = TRUE;
			}			
		}

		if (psl->bNonBlocking){
			nbmode = TRUE;
			gft.ioctlsocket(s, FIONBIO, &nbmode);
		}

		if (psl->hAsyncWnd)
			gft.WSAAsyncSelect(s, psl->hAsyncWnd, psl->uAsyncMsg, psl->lAsyncEvent);
	}

	if (psl && psl->STelChk.bSSLActive && !psl->STelChk.bInSSLCall){
x_OutputDebugString("S ");

		psl->STelChk.bInSSLCall = TRUE;
		ret = SSL_write(psl->STelChk.con, (char *)buf, len);
		psl->STelChk.bInSSLCall = FALSE;

	} else {
x_OutputDebugString("O ");

		ret = gft.send(s, buf, len, flags);

	}	

	return ret;
}


//														SSLftp
/////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////
//	STelSock


/////////////////////////////////////////////////////////////////
//
//	accept
//
SOCKET PASCAL FAR EXPORT
accept (SOCKET s, struct sockaddr FAR *addr, int FAR *addrlen)
{
	SOCKET	ret;
	struct SocketList *psls, *pslret;

x_OutputDebugString("accept ");

	switch(giNegotiation){
	case SSLFTP:
		ret = SSLftp_accept(s, addr, addrlen);
		break;

	default:
		ret = gft.accept(s, addr, addrlen);
		if (ret != SOCKET_ERROR)
			SocketListAddTail(ret);
		break;
	}

	if (ret != SOCKET_ERROR){
		psls   = SocketListGetPos(s);
		pslret = SocketListGetPos(ret);

		// Inherit the socket mode
		pslret->bNonBlocking = psls->bNonBlocking;
	}

	return ret;
}


/////////////////////////////////////////////////////////////////
//
//	closesocket
//		Free the context of SSL telnet
//
int PASCAL FAR EXPORT
closesocket (SOCKET s)
{
	int		ret;
	struct SocketList *psl;

x_OutputDebugString("\nclosesocket ");

	ret = gft.closesocket(s);

	switch(giNegotiation){
	case SSLFTP:
		if (gsFtpCtrl == s){
			gsFtpCtrl    = INVALID_SOCKET;
			gsFtpData    = INVALID_SOCKET;
			gbFtpChecked = FALSE;
		}
		break;

	case SSLTELNET:
	case SSLONLY:
	default:
		break;
	}

	// The release of the socket is delayed for a bit.
	while (psl = SocketListGetPosByStat(SS_CLOSED)){
		SocketListDelete(psl->s);
	}
	psl = SocketListGetPos(s);
	if (psl){
		psl->iStat = SS_CLOSED;
	}

	return ret;
}


/////////////////////////////////////////////////////////////////
//
//	connect
//
int PASCAL FAR EXPORT
connect (SOCKET s, const struct sockaddr FAR *name, int namelen)
{
	int		ret;

	switch (giNegotiation){
	case SSLFTP:
		if (gsFtpCtrl  == INVALID_SOCKET &&
			name->sa_family == AF_INET &&
			IPPORT_FTP == ntohs(((struct sockaddr_in *)name)->sin_port)){
			gsFtpCtrl = s;
		}
		ret = gft.connect(s, name, namelen);
		// PASV mode is Not Yet Implemtnted.
		break;

	case SSLONLY:
		if (name->sa_family == AF_INET){
			switch (ntohs(((struct sockaddr_in *)name)->sin_port)){
			case IPPORT_SMTP:
			case 110:
				ret = gft.connect(s, name, namelen);
				break;
			default:
				ret = SSLtelnet_connect(s, name, namelen);
				break;
			}

		} else {
			ret = SOCKET_ERROR;
			gft.WSASetLastError(WSAEAFNOSUPPORT);
		}
		break;

	case SSLTELNET:
	default:
		ret = SSLtelnet_connect(s, name, namelen);
		break;
	}

	return ret;
}


/////////////////////////////////////////////////////////////////
//
//	ioctlsocket
//
int PASCAL FAR EXPORT
ioctlsocket (SOCKET s, long cmd, u_long FAR *argp)
{
	int		ret;
	struct SocketList	*psl;

	psl = SocketListGetPos(s);

x_OutputDebugString("ioctlsocket ");
	
	ret = gft.ioctlsocket(s, cmd, argp);

	if (ret != SOCKET_ERROR){
		switch(cmd){
		case FIONBIO:
			if (psl && argp){
				if (*argp)
					psl->bNonBlocking = TRUE;
				else
					psl->bNonBlocking = FALSE;
			}
			break;
		default:
			break;
		}
	}
	
	return ret;
}


/////////////////////////////////////////////////////////////////
//
//	recv
//
int PASCAL FAR EXPORT
recv (SOCKET s, char FAR * buf, int len, int flags)
{
	int		ret;
	struct SocketList	*psl;

	psl = SocketListGetPos(s);

x_OutputDebugString("recv ");

	if (psl && psl->STelChk.bSSLActive && !psl->STelChk.bInSSLCall){
x_OutputDebugString("S ");

		psl->STelChk.bInSSLCall = TRUE;
		ret = SSL_read(psl->STelChk.con, buf, len);
		psl->STelChk.bInSSLCall = FALSE;

	} else {
x_OutputDebugString("O ");

		ret = gft.recv(s, buf, len, flags);

	}

	return ret;
}


/////////////////////////////////////////////////////////////////
//
//	send
//
int PASCAL FAR EXPORT
send (SOCKET s, const char FAR * buf, int len,
    int flags)
{
	int		ret;

	switch(giNegotiation){
	case SSLFTP:
		ret = SSLftp_send(s, buf, len, flags);
		break;

	case SSLTELNET:
	case SSLONLY:
	default:
		ret = SSLtelnet_send(s, buf, len, flags);
		break;
	}

	return ret;
}


///////////////////////////////////////////////////////////////
//
//	socket
//
SOCKET PASCAL FAR EXPORT
socket (int af, int type, int protocol)
{
	SOCKET s;

x_OutputDebugString("socket ");

	s = gft.socket(af, type, protocol);

	if (SocketListGetPos(s))
		SocketListDelete(s);

	SocketListAddTail(s);

	return s;
}


///////////////////////////////////////////////////////////////
//
//	WSAAsyncSelect
//
int PASCAL FAR EXPORT
WSAAsyncSelect(SOCKET s, HWND hWnd, u_int wMsg, long lEvent)
{
	int		ret;
	struct SocketList	*psl;

	psl = SocketListGetPos(s);

x_OutputDebugString("WSAAsyncSelect ");

	ret = gft.WSAAsyncSelect(s, hWnd, wMsg, lEvent);

	if (psl){
		psl->bNonBlocking = TRUE;

		psl->hAsyncWnd    = hWnd;
		psl->uAsyncMsg    = wMsg;
		psl->lAsyncEvent  = lEvent;
	}

	return ret;
}


///////////////////////////////////////////////////////////////
//
//	WSAStartup
//
int PASCAL FAR EXPORT
WSAStartup(WORD wVersionRequired, LPWSADATA lpWSAData)
{
	int ret;

x_OutputDebugString("WSAStartup ");

	if (gft.hWinsock == NULL){
		if (LoadWinsock() == FALSE){
			return WSASYSNOTREADY;
		}
	}

	ret = gft.WSAStartup(wVersionRequired, lpWSAData);

	return ret;
}

/////////////////////////////////////////////////////////////////////
//
//	Set_SSLCTX
//
int Set_SSLCTX(SSL_CTX *ctx)
{
	char *CertFile = NULL, *KeyFile = NULL, *CAfile = NULL, *CApath = NULL;

	if (gCertFile[0])
		CertFile = gCertFile;
	if (gKeyFile[0])
		KeyFile = gKeyFile;
	if (gCAFile[0])
		CAfile = gCAFile;
	if (gCAPath[0])
		CApath = gCAPath;

	SSL_CTX_set_verify(ctx, gVerifyDepth ? SSL_VERIFY_PEER : SSL_VERIFY_NONE,
						Verify_CB);

	if (!gNotUseCert && CertFile != NULL){
		if (SSL_CTX_use_certificate_file(ctx, CertFile, SSL_FILETYPE_PEM) <= 0)
			return FALSE;

		if (KeyFile == NULL)
			KeyFile = CertFile;

		SSL_CTX_set_default_passwd_cb(ctx, getkeycb);
		if (SSL_CTX_use_PrivateKey_file(ctx, KeyFile, SSL_FILETYPE_PEM) <= 0)
			return FALSE;
	}

#ifdef _SSLEAY06_
	if ((!SSL_load_verify_locations(ctx, CAfile, CApath)) ||
		(!SSL_set_default_verify_paths(ctx))){
#else
	if ((!SSL_CTX_load_verify_locations(ctx, CAfile, CApath)) ||
		(!SSL_CTX_set_default_verify_paths(ctx))){
#endif /* _SSLEAY06_ */
		return FALSE;
	}

	return TRUE;
}


/////////////////////////////////////////////////////////////////////
//
//	getkeycb
//
#ifdef _SSLEAY06_
int getkeycb(char *buf, int len)
#else
int getkeycb(char *buf, int len, int w)
#endif /* _SSLEAY06_ */
{
	char pass[PASS_PHRASE_MAX_LEN];
	int  res;

	if (DialogBoxParam(ghModule, MAKEINTRESOURCE(IDD_PASSPHRASE),
					NULL, PassphraseDlgProc, (LPARAM)pass)){
		res = strlen(pass);
		if (len < res)
			return -1;

		strncpy(buf, pass, res);
		memset(pass, 0, PASS_PHRASE_MAX_LEN);
		return res;

	} else {
		return -1;
	}
}


//////////////////////////////////////////////////////////////
//
//	Passphrase Input Dialog Procedure
//
BOOL CALLBACK PassphraseDlgProc(HWND hWnd, UINT uMessage,
				WPARAM wParam, LPARAM lParam)
{
	static char *pass;

    switch(uMessage) {
    case WM_INITDIALOG:
	{
		POINT pt, spt;
		RECT  r;

		pass = (char *)lParam;

		GetCursorPos(&pt);
		GetWindowRect(hWnd, &r);
		spt.x = GetSystemMetrics(SM_CXSCREEN) - pt.x - (r.right - r.left);
		spt.y = GetSystemMetrics(SM_CYSCREEN) - pt.y - (r.bottom - r.top);
		if (spt.x < 0) pt.x += spt.x;
		if (spt.y < 0) pt.y += spt.y;

		MoveWindow(hWnd, pt.x, pt.y, r.right - r.left, r.bottom - r.top, TRUE);
		ShowWindow(hWnd, SW_SHOWNORMAL);
		SetForegroundWindow(hWnd);
	}
		break;

	case WM_COMMAND:
		switch(wParam){
		case IDOK:
			GetDlgItemText(hWnd, IDC_PASS, pass, PASS_PHRASE_MAX_LEN);
			EndDialog(hWnd, TRUE);
			break;

		case IDCANCEL:
			EndDialog(hWnd, FALSE);
			break;

		default:
			return FALSE;
		}
		break;

	default:
		return FALSE;
	}
	return TRUE;
}																	   	


//////////////////////////////////////////////////////////////
//
//	Verify Callback
//
#ifdef _SSLEAY06_
int Verify_CB(int ok, X509 *xs, X509 *xi, int depth, int error, char *arg)
#else
int Verify_CB(int ok, X509_STORE_CTX *ctx)
#endif /* _SSLEAY06_ */
{
	char dbuf[256];

#ifndef _SSLEAY06_
	int  error, depth;
	error = X509_STORE_CTX_get_error(ctx);
	depth = X509_STORE_CTX_get_error_depth(ctx);
#endif /* _SSLEAY06_ */

	if (!ok){
		wsprintf(dbuf, "verify error: %s",
#ifdef _SSLEAY06_
				X509_cert_verify_error_string(error));
#else
				X509_verify_cert_error_string(error));
#endif /* SSLEAY06_ */
		ErrorMsgBox(dbuf);

#ifdef _SSLEAY06_
		if (gVerifyDepth <= depth){
#else
		if (gVerifyDepth >= depth) {
#endif /* SSLEAY06_ */
			ok = 1;
		} else {
			ok = 0;
		}
	}

	return(ok);
}


//////////////////////////////////////////////////////////////
//
//	Error Message Dialog
//
int ErrorMsgBox(char *errmsg)
{
	if (gErrMsgTime)
		DialogBoxParam(ghModule, MAKEINTRESOURCE(IDD_ERRMSG),
						NULL, ErrmsgDlgProc, (LPARAM)errmsg);
	return TRUE;
}


//////////////////////////////////////////////////////////////
//
//	Error Message Dialog
//
BOOL CALLBACK ErrmsgDlgProc(HWND hWnd, UINT uMessage,
				WPARAM wParam, LPARAM lParam)
{
    switch(uMessage) {
    case WM_INITDIALOG:
		SetWindowText(GetDlgItem(hWnd, IDC_ERRMSG), (char *)lParam);
		SetTimer(hWnd, IDT_ERRMSG, gErrMsgTime * 1000, NULL);
		break;

	case WM_TIMER:
		if (wParam == IDT_ERRMSG){
			KillTimer(hWnd, IDT_ERRMSG);
			EndDialog(hWnd, TRUE);
		}
		break;

	default:
		return FALSE;
	}
	return TRUE;
}																	   	


/////////////////////////////////////////////////////////////////////
//
//	GetOptions
//		This function determines the negotiation mode for the
//		applcations. The mode is set on the registory.
//
int GetOptions()
{
	DWORD dwCliStat;
	DWORD dwMode;

	dwCliStat = GetClientStat();

	// Get negotiation mode
	dwMode = (NEGO_MODE & dwCliStat);
	switch( dwMode ){
	case SSLFTP:
		gsFtpCtrl     = INVALID_SOCKET;
		gsFtpData     = INVALID_SOCKET;
		gbFtpChecked  = FALSE;
	case SSLTELNET:
	case SSLONLY:
		giNegotiation = dwMode;
		break;

	default:
		giNegotiation = SSLTELNET;
		break;
	}

	// Get verify mode
	dwMode = (VERIFY_DEPTH & dwCliStat);
	gVerifyDepth = dwMode >> 8;

	// Use/Not Use certification
	dwMode = (NOTUSE_CERT & dwCliStat);
	gNotUseCert = dwMode >> 12;

	// Get error message display time
	dwMode = (ERRMSG_TIME & dwCliStat);
	gErrMsgTime = dwMode >> 16;

	return TRUE;
}


/////////////////////////////////////////////////////////////////////
//
//	DllMain
//
BOOL APIENTRY DllMain(HANDLE hModule, DWORD ul_reason_for_call, LPVOID lpReserved)
{
	struct LocalData		*ld;

	switch(ul_reason_for_call){
	case DLL_PROCESS_ATTACH:
		ghModule = hModule;

		TlsIndex = TlsAlloc();

		gft.hWinsock  = NULL;
		giNegotiation = SSLTELNET;

		ld = (struct LocalData *)LocalAlloc(LPTR, sizeof(struct LocalData));
		TlsSetValue(TlsIndex, (LPVOID)ld);
		ld->hWinsock = gft.hWinsock;

		GetOptions();
		GetPEMFiles();
		break;

	case DLL_THREAD_ATTACH:
		ld = (struct LocalData *)LocalAlloc(LPTR, sizeof(struct LocalData));
		TlsSetValue(TlsIndex, (LPVOID)ld);

		ld->hWinsock = LoadLibrary(ORIGINAL_WSOCK32);
		break;

	case DLL_THREAD_DETACH:
		ld = (struct LocalData *)TlsGetValue(TlsIndex);

		FreeLibrary(ld->hWinsock);

		LocalFree(ld);
		break;

	case DLL_PROCESS_DETACH:
		ld = (struct LocalData *)TlsGetValue(TlsIndex);
		LocalFree(ld);

		FreeWinsock();

		TlsFree(TlsIndex);
		break;
	}
	return TRUE;
}


/////////////////////////////////////////////////////////////////////////
//
//	LoadWinsock
//
int LoadWinsock()
{
	gft.hWinsock = LoadLibrary(ORIGINAL_WSOCK32);

	if (gft.hWinsock == NULL){
		char szWSPath[MAX_PATH];

		if (!GetSystemDirectory(szWSPath, MAX_PATH))
			return FALSE;
		
		strcat(szWSPath, "\\wsock32.dll");

		gft.hWinsock = LoadLibrary(szWSPath);

	}

	if (gft.hWinsock == NULL)
		return FALSE;

	gft.accept                = (void *)GetProcAddress(gft.hWinsock, "accept");
	gft.bind                  = (void *)GetProcAddress(gft.hWinsock, "bind");
	gft.closesocket           = (void *)GetProcAddress(gft.hWinsock, "closesocket");
	gft.connect               = (void *)GetProcAddress(gft.hWinsock, "connect");
	gft.getpeername           = (void *)GetProcAddress(gft.hWinsock, "getpeername");
	gft.getsockname           = (void *)GetProcAddress(gft.hWinsock, "getsockname");
	gft.getsockopt            = (void *)GetProcAddress(gft.hWinsock, "getsockopt");
	gft.htonl                 = (void *)GetProcAddress(gft.hWinsock, "htonl");
	gft.htons                 = (void *)GetProcAddress(gft.hWinsock, "htons");
	gft.inet_addr             = (void *)GetProcAddress(gft.hWinsock, "inet_addr");
	gft.inet_ntoa             = (void *)GetProcAddress(gft.hWinsock, "inet_ntoa");
	gft.ioctlsocket           = (void *)GetProcAddress(gft.hWinsock, "ioctlsocket");
	gft.listen                = (void *)GetProcAddress(gft.hWinsock, "listen");
	gft.ntohl                 = (void *)GetProcAddress(gft.hWinsock, "ntohl");
	gft.ntohs                 = (void *)GetProcAddress(gft.hWinsock, "ntohs");
	gft.recv                  = (void *)GetProcAddress(gft.hWinsock, "recv");
	gft.recvfrom              = (void *)GetProcAddress(gft.hWinsock, "recvfrom");
	gft.select                = (void *)GetProcAddress(gft.hWinsock, "select");
	gft.send                  = (void *)GetProcAddress(gft.hWinsock, "send");
	gft.sendto                = (void *)GetProcAddress(gft.hWinsock, "sendto");
	gft.setsockopt            = (void *)GetProcAddress(gft.hWinsock, "setsockopt");
	gft.socket                = (void *)GetProcAddress(gft.hWinsock, "socket");
	gft.shutdown              = (void *)GetProcAddress(gft.hWinsock, "shutdown");

	gft.gethostbyaddr         = (void *)GetProcAddress(gft.hWinsock, "gethostbyaddr");
	gft.gethostbyname         = (void *)GetProcAddress(gft.hWinsock, "gethostbyname");
	gft.getprotobyname        = (void *)GetProcAddress(gft.hWinsock, "getprotobyname");
	gft.getprotobynumber      = (void *)GetProcAddress(gft.hWinsock, "getprotobynumber");
	gft.getservbyname         = (void *)GetProcAddress(gft.hWinsock, "getservbyname");
	gft.getservbyport         = (void *)GetProcAddress(gft.hWinsock, "getservbyport");
	gft.gethostname           = (void *)GetProcAddress(gft.hWinsock, "gethostname");

	gft.WSAAsyncSelect        = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncSelect");
	gft.WSAAsyncGetHostByAddr = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncGetHostByAddr");
	gft.WSAAsyncGetHostByName = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncGetHostByName");
	gft.WSAAsyncGetProtoByNumber = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncGetProtoByNumber");
	gft.WSAAsyncGetProtoByName = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncGetProtoByName");
	gft.WSAAsyncGetServByPort = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncGetServByPort");
	gft.WSAAsyncGetServByName = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncGetServByName");
	gft.WSACancelAsyncRequest = (void *)GetProcAddress(gft.hWinsock, "WSACancelAsyncRequest");
	gft.WSASetBlockingHook    = (void *)GetProcAddress(gft.hWinsock, "WSASetBlockingHook");
	gft.WSAUnhookBlockingHook = (void *)GetProcAddress(gft.hWinsock, "WSAUnhookBlockingHook");
	gft.WSAGetLastError       = (void *)GetProcAddress(gft.hWinsock, "WSAGetLastError");
	gft.WSASetLastError       = (void *)GetProcAddress(gft.hWinsock, "WSASetLastError");
	gft.WSACancelBlockingCall = (void *)GetProcAddress(gft.hWinsock, "WSACancelBlockingCall");
	gft.WSAIsBlocking         = (void *)GetProcAddress(gft.hWinsock, "WSAIsBlocking");
	gft.WSAStartup            = (void *)GetProcAddress(gft.hWinsock, "WSAStartup");
	gft.WSACleanup            = (void *)GetProcAddress(gft.hWinsock, "WSACleanup");

	gft.__WSAFDIsSet          = (void *)GetProcAddress(gft.hWinsock, "__WSAFDIsSet");

	gft.EnumProtocolsA		  = (void *)GetProcAddress(gft.hWinsock, "EnumProtocolsA");
	gft.EnumProtocolsW		  = (void *)GetProcAddress(gft.hWinsock, "EnumProtocolsW");
	gft.GetAddressByNameA	  = (void *)GetProcAddress(gft.hWinsock, "GetAddressByNameA");
	gft.GetAddressByNameW	  = (void *)GetProcAddress(gft.hWinsock, "GetAddressByNameW");
	gft.GetNameByTypeA		  = (void *)GetProcAddress(gft.hWinsock, "GetNameByTypeA");
	gft.GetNameByTypeW		  = (void *)GetProcAddress(gft.hWinsock, "GetNameByTypeW");
	gft.GetServiceA			  = (void *)GetProcAddress(gft.hWinsock, "GetServiceA");
	gft.GetServiceW			  = (void *)GetProcAddress(gft.hWinsock, "GetServiceW");
	gft.GetTypeByNameA		  = (void *)GetProcAddress(gft.hWinsock, "GetTypeByNameA");
	gft.GetTypeByNameW		  = (void *)GetProcAddress(gft.hWinsock, "GetTypeByNameW");
	gft.SetServiceA			  = (void *)GetProcAddress(gft.hWinsock, "SetServiceA");
	gft.SetServiceW			  = (void *)GetProcAddress(gft.hWinsock, "SetServiceW");
	gft.TransmitFile		  = (void *)GetProcAddress(gft.hWinsock, "TransmitFile");
	gft.WSARecvEx			  = (void *)GetProcAddress(gft.hWinsock, "WSARecvEx");

	gft.WsControl             = (void *)GetProcAddress(gft.hWinsock, "WsControl");
	gft.closesockinfo         = (void *)GetProcAddress(gft.hWinsock, "closesockinfo");
	gft.Arecv                 = (void *)GetProcAddress(gft.hWinsock, "Arecv");
	gft.Asend                 = (void *)GetProcAddress(gft.hWinsock, "Asend");
	gft.WSHEnumProtocols      = (void *)GetProcAddress(gft.hWinsock, "WSHEnumProtocols");
	gft.inet_network          = (void *)GetProcAddress(gft.hWinsock, "inet_network");
	gft.getnetbyname          = (void *)GetProcAddress(gft.hWinsock, "getnetbyname");
	gft.rcmd                  = (void *)GetProcAddress(gft.hWinsock, "rcmd");
	gft.rexec                 = (void *)GetProcAddress(gft.hWinsock, "rexec");
	gft.rresvport             = (void *)GetProcAddress(gft.hWinsock, "rresvport");
	gft.sethostname           = (void *)GetProcAddress(gft.hWinsock, "sethostname");
	gft.dn_expand             = (void *)GetProcAddress(gft.hWinsock, "dn_expand");
	gft.s_perror              = (void *)GetProcAddress(gft.hWinsock, "s_perror");
	gft.NPLoadNameSpaces      = (void *)GetProcAddress(gft.hWinsock, "NPLoadNameSpaces");

x_OutputDebugString("END: LoadWinsock\n");

	return TRUE;
}


///////////////////////////////////////////////////////////////
//
//	FreeWinsock
//
int FreeWinsock()
{
	struct SocketList	*psl;

	while (psl = SocketListGetPosByStat(SS_CLOSED)){
		SocketListDelete(psl->s);
	}

	FreeLibrary(gft.hWinsock);

x_OutputDebugString("FreeWinsock\n");

	return TRUE;
}
